import math
import xml.etree.ElementTree as ET
from dataclasses import dataclass, InitVar
from typing import Any, Dict, List, Set, Union, Optional, Tuple
from collections import defaultdict
import pprint
from fractions import Fraction
import argparse
pp = pprint.PrettyPrinter(indent=2)
@dataclass(frozen=True)
class Values:
	""""""
	id: int
	values: Dict[str, str]

@dataclass(frozen=True)
class Node:
	""""""
	state: Values
	combinatorial: Optional[Values]
	input: Optional[Values]
	

@dataclass(frozen=True)
class Counterexample:
	""""""
	type: int
	id: int
	descr: list
	nodes: List[Node]
	
def parse_values(value_holder: ET) -> Values:
	""" Extract all variables and respective value one XML tag deep from
	value_holder parameter and returns them as a Values object."""

	values = {}
	if value_holder:
		for value in value_holder.findall("value") or []:
			values[value.attrib["variable"]] = value.text

	return Values(value_holder.attrib["id"], values) if value_holder else None


def parse_counterexample(counterexample_xml: str) -> Counterexample:
	""" Parses a counterexmple produced by nuXmv.

	Parameters
	----------
	counterexample_xml
		The filename of an xml produced by nuXmv containing a counterexample of
		a model.

	Returns
	-------
	A Counterexample dataclass containing the parsed data"""

	counterexample_et = ET.parse(counterexample_xml)
	root_node = counterexample_et.getroot()

	# Iterate over nodes
	parsed_nodes = []
	for node in root_node.findall("node"):
		value_holders = []
		for tag in ["state", "combinatorial", "input"]:
			value_holders.append(parse_values(node.find(tag)))

		parsed_nodes.append(Node(value_holders[0], value_holders[1], value_holders[2]))

	# Create the final datastructure
	ret = Counterexample(root_node.attrib["type"], root_node.attrib["id"],
			root_node.attrib["desc"], parsed_nodes)
	sim_steps = len(parsed_nodes)-1
	return ret, sim_steps


def removesuffix(var, suff):
	if var.endswith(suff):
		var = var[:-len(suff)]
	return var


def removeprefix(var, pref):
	if var.startswith(pref):
		var = var[len(pref):]
	return var


def get_float(s):
	if "f'" in s:
		s = s.replace("f'","")
		res = float(Fraction(s))
	else:
		res = float(s)
	return res

class ParamsParser:
	def is_active_flag(self, key, val):
		if "agent_" in key and ".me." not in key and "is_" not in key and "bootstrap" not in key and "TRUE" in val:
			car = removesuffix(key, ".active")
			return removeprefix(car, "agent_")
		return None

	def get_lane(self, node, car):
		lane = node.state.values["agent_"+car+".lane"]
		return -(int(lane)+1)

	def get_x_pos(self, node, car):
		x_pos = node.state.values["agent_"+car+".pos"]
		return get_float(x_pos)

	def get_acceleration(self, node, car):
		acceleration = node.input.values["agent_"+car+".acceleration"]
		return get_float(acceleration)

	def get_speed(self, node, car):
		speed = node.state.values["agent_"+car+".speed"]
		return get_float(speed)

	def get_time_step(self, node):
		duration = node.state.values["simulation.time_step"]
		return get_float(duration)
	
	def get_target_lane(self, node, car):
		if car == "ego":
			target_lane = node.combinatorial.values["agent_"+car+".me.target_lane"]
		else:
			target_lane = node.combinatorial.values["agent_"+car+".target_lane"]
		return -(int(target_lane)+1)

	def get_weather(self, node):
		weather = node.state.values["weather"]
		return weather 
# end of class ParamsParser


def create_dict(counterexample, parser):
	# from the first node
	# active cars in the environment     
	cars_act=[]
	node=counterexample.nodes[0]
	for key, value in node.state.values.items():
		car = parser.is_active_flag(key, value)
		if car is not None:
			cars_act.append(car)

	# init parameters for agents
	params_g = defaultdict(dict)
	for car in cars_act:
		lane = parser.get_lane(node, car)
		x_pos = parser.get_x_pos(node, car)
		params_g[car] = {"initial": defaultdict(list),
						 car+"behavior": defaultdict(list)}
		params_g[car]['initial']['lane'] = lane
		params_g[car]['initial']['x_pos'] = x_pos

	for car in cars_act:
		if car != "ego":
			if params_g[car]['initial']['lane'] != params_g['ego']['initial']['lane']:
				dist = \
					math.sqrt((params_g[car]['initial']['x_pos'] - \
					params_g['ego']['initial']['x_pos'])**2 + 3.5**2)
				if params_g[car]['initial']['x_pos'] > \
					params_g['ego']['initial']['x_pos']:
					dist = -dist
				params_g[car]['initial']['rel_dist_to_ego'] = dist
			else:
				params_g[car]['initial']['rel_dist_to_ego'] = \
				params_g[car]['initial']['x_pos'] - \
				params_g['ego']['initial']['x_pos']

	ego_pos = params_g['ego']['initial']['x_pos']
	last_ego_pos = ego_pos
	
	# iterate over all nodes to extract behaviors and corresponding parameters
	for state in counterexample.nodes:
		if state.input:
			for car in cars_act:
				acceleration = parser.get_acceleration(state, car)
				if acceleration >= 0:
					throttle_in = acceleration/5.6
				else:
					throttle_in = acceleration/4.6
				lane = parser.get_lane(state, car)
				duration = parser.get_time_step(state)
				
				speed = parser.get_speed(state, car)
				x_pos = parser.get_x_pos(state, car)
				target_lane = parser.get_target_lane(state, car)
				
				params_g[car][car+"behavior"]['throttle_in'].append(throttle_in)
				params_g[car][car+"behavior"]['lanetoFollow'].append(lane)
				params_g[car][car+"behavior"]['duration'].append(duration)
				
				params_g[car][car+"behavior"]['speed'].append(speed)
				params_g[car][car+"behavior"]['x_pos'].append(x_pos)
				params_g[car][car+"behavior"]['target_lane'].append(target_lane)
				params_g[car][car+"behavior"]['acceleration'].append(acceleration)
		last_ego_pos = parser.get_x_pos(state, 'ego')

	ego_distance = last_ego_pos - ego_pos
	params_g['ego_distance'] = ego_distance

	for state in counterexample.nodes:
		if not state.input:
			for car in cars_act:
				lane = parser.get_lane(state, car)
				duration = parser.get_time_step(state)
				speed = parser.get_speed(state, car)
				x_pos = parser.get_x_pos(state, car)
				
				params_g[car][car+"behavior"]['lanetoFollow'].append(lane)
				params_g[car][car+"behavior"]['duration'].append(duration)
				params_g[car][car+"behavior"]['speed'].append(speed)
				params_g[car][car+"behavior"]['x_pos'].append(x_pos)



	for i in range(len(params_g['ego']['egobehavior']['x_pos'])-1):
		for car in cars_act:
			if car != "ego":
				dist_travelled = params_g[car][car+"behavior"]['x_pos'][i] - params_g[car][car+"behavior"]['x_pos'][i+1]
				params_g[car][car+"behavior"]['dist_step'].append(dist_travelled)
				
	for i in range(len(params_g['ego']['egobehavior']['x_pos'])):
		for car in cars_act:
			if car != "ego":
				rel_dist_to_ego = params_g[car][car+"behavior"]['x_pos'][i] - params_g['ego']['egobehavior']['x_pos'][i]
				params_g[car][car+"behavior"]['rel_dist_to_ego'].append(rel_dist_to_ego)
	return params_g


def trace(path_dir_xml, parser=None):
	counterexample, n_sim_steps = parse_counterexample(path_dir_xml)
	if parser is None:
		parser = ParamsParser()
	counterexample_dict = create_dict(counterexample, parser)
	myKeys = list(counterexample_dict.keys())
	myKeys.sort()
	sorted_dict = {i: counterexample_dict[i] for i in myKeys}
	return sorted_dict #, n_sim_steps

def indent(elem, level=0):
            """
            Add whitespace indentation to an XML element and its subelements.
            """
            i = "\n" + level * "  "
            if len(elem):
                if not elem.text or not elem.text.strip():
                    elem.text = i + "  "
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
                for elem in elem:
                    indent(elem, level + 1)
                if not elem.tail or not elem.tail.strip():
                    elem.tail = i
            else:
                if level and (not elem.tail or not elem.tail.strip()):
                    elem.tail = i

def create_ego_route(path: str, path_dir_xml, weather: str, parser=None): # for concrete trace
#def create_ego_route(path: str, path_dir_xml, parser=None):
	counterexample, n_sim_steps = parse_counterexample(path_dir_xml)
	if parser is None:
		parser = ParamsParser()
	#weather = parser.get_weather(counterexample.nodes[0]) # from abstract trace

	CE = ET.Element("routes")

	# create sub element
	CE = ET.SubElement(CE, "routes")
	#CE.attrib = {"desc": "Simulator Execution"}

	#CE.attrib = {"weather": weather}

	node = ET.SubElement(CE, "route")
	node.attrib = {"id": "0", "town": "Town06"}
	state = ET.SubElement(node, "weather")
	if weather == "ClearNoon":
		state.attrib = {"id":"ClearNoon", "cloudiness":"0.000000", "precipitation":"0.000000", "precipitation_deposits":"0.000000", "wind_intensity":"0.000000", "sun_azimuth_angle":"90.000000", "sun_altitude_angle":"90.000000", "fog_density":"0.000000", "fog_distance":"0.000000", "fog_falloff":"0.000000", "wetness":"0.000000"}
	if weather ==  "MidRainNoon":
		state.attrib = {"id":"MidRainNoon", "cloudiness":"45.000000", "precipitation":"50.000000", "precipitation_deposits":"50.000000", "wind_intensity":"1.000000", "sun_azimuth_angle":"90.000000", "sun_altitude_angle":"70.000000", "fog_density":"0.000000", "fog_distance":"0.000000", "fog_falloff":"0.000000", "wetness":"0.000000"}
	elif weather == "HardRainNight":
		state.attrib = {"id":"HardRainNight", "cloudiness":"90.000000", "precipitation":"100.000000", "precipitation_deposits":"100.000000", "wind_intensity":"1.000000", "sun_azimuth_angle":"225.000000", "sun_altitude_angle":"-80.000000", "fog_density":"0.000000", "fog_distance":"0.000000", "fog_falloff":"0.000000", "wetness":"0.000000"}
	elif weather == "Sunset":
		state.attrib = {"id":"Sunset", "cloudiness":"0.000000", "precipitation":"0.000000", "precipitation_deposits":"0.000000", "wind_intensity":"0.400000", "sun_azimuth_angle":"180.000000", "sun_altitude_angle":"20.000000", "fog_density":"0.000000", "fog_distance":"0.000000", "fog_falloff":"0.000000", "wetness":"0.000000"}
	elif weather == "HardRainTwilight":
		state.attrib = {"id":"HardRainTwilight", "cloudiness":"90.000000", "precipitation":"70.000000", "precipitation_deposits":"100.000000", "wind_intensity":"1.000000", "sun_azimuth_angle":"315.000000", "sun_altitude_angle":"0.000000", "fog_density":"0.000000", "fog_distance":"0.000000", "fog_falloff":"0.000000", "wetness":"0.000000"}
	elif weather == "WetMorning":
		state.attrib = {"id":"WetMorning", "cloudiness":"5.000000", "precipitation":"0.000000", "precipitation_deposits":"80.000000", "wind_intensity":"0.350000", "sun_azimuth_angle":"0.000000", "sun_altitude_angle":"10.000000", "wetness":"30.000000"}
	
	way1 = ET.SubElement(node, "waypoint")
	way1.attrib = {"pitch" : "0", "roll" : "0", "x" : "560.3289184570312", "y" : "-18.5191191482543945", "yaw" : "180.41941833496094", "z" : "0"}

	way2 = ET.SubElement(node, "waypoint")
	way2.attrib = {"pitch" : "0", "roll" : "0", "x" : "360.858123779296", "y" : "-18.514790267944336", "yaw" : "179.79132080078125", "z" : "0"}

	# create a new XML file with the results
	tree = ET.ElementTree(CE)
	indent(CE)
	ego_route_file = path
	with open(ego_route_file, "wb") as fh:
		tree.write(fh, encoding ='UTF-8', xml_declaration = True)

def main():
	p = argparse.ArgumentParser()
	p.add_argument('-o', '--output')
	p.add_argument('input')
	opts = p.parse_args()
	d, n = trace(opts.input)
	if opts.output:
		with open(opts.output, 'w') as out:
			pp.pprint(d, out)
			pp.pprint(n, out)
	else:
		pp.pprint(d)
		pp.pprint(n)


if __name__ == '__main__':
	main()

