import os
import sys
# make sure the bundled version of scenic is found
sys.path.insert(0, os.path.join(os.path.dirname(__file__),
                                'simulator', 'scenic'))

from absgen import AbstractGenerator
from concretizer import RoverConcretizer, ScenicConcretizer, CarlaConcretizer
from simulator import RoverSimulator, ScenicSimulator, CarlaSimulator
from monitor import RoverMissionMonitor, ScenicMonitor, CarlaMonitor
from concretizer import RoverConcretizer, ScenicConcretizer
from monitor import RoverMissionMonitor, ScenicMonitor
from coverage import CoverageManager
from abc import abstractmethod


class VIVASConfiguration:
    def __init__(self, settings, output_dir):
        self.settings = settings
        self.output_dir = output_dir

    def get_abstract_generator(self):
        ag_set = self.settings["absgen_settings"]
        model = ag_set['model']
        if not os.path.isabs(model):
            model = os.path.abspath(
                os.path.join(self.settings['basedir'], model))
        return AbstractGenerator(
            model,
            ag_set.get("path_to_exe"),
            pythonpath=ag_set.get("python_paths"),
            solver=ag_set.get("solver"),
            output_dir=os.path.join(self.output_dir, 'absgen'),
            solver_logic=ag_set.get("solver_logic"),
            timeout=ag_set["timeout"])

    def get_coverage_manager(self):
        pred_map_file = None
        cov = self.settings.get("coverage_settings")
        if cov:
            pred_map_file = cov.get("predicate_map")
            if pred_map_file and not os.path.isabs(pred_map_file):
                pred_map_file = os.path.abspath(
                    os.path.join(self.settings['basedir'], pred_map_file))
        return CoverageManager(pred_map_file, self.output_dir)

    @abstractmethod
    def get_concretizer(self):
        pass

    @abstractmethod
    def get_simulator(self):
        pass

    @abstractmethod
    def get_monitor(self):
        pass

# end of class VIVASConfiguration


class RoverConfiguration(VIVASConfiguration):
    def get_concretizer(self):
        cn_set = self.settings["concretizer_settings"]
        return RoverConcretizer(
            output_dir=os.path.join(self.output_dir, 'concretizer'),
            num_samples=cn_set.get('scenarios_to_generate', 1),
            default_full_goalset=cn_set.get('complete_plan', True))

    def get_simulator(self):
        sim_set = self.settings["simulator_settings"]

        return RoverSimulator(
            sim_set["simulator_path"],
            sim_set["operations_path"],
            ld_library_path=sim_set["ld_lib_path"],
            release_home=sim_set["release_home"],
            output_dir=os.path.join(self.output_dir, 'simulator'),
            novelty_detection=sim_set.get('novelty_detection'),
            novelty_replanning=sim_set.get('novelty_replanning', False))

    def get_monitor(self):
        return RoverMissionMonitor(os.path.join(self.output_dir, 'monitor'))

# end of class RoverConfiguration


class ScenicConfiguration(VIVASConfiguration):
    def get_concretizer(self):
        return ScenicConcretizer(
            output_dir=os.path.join(self.output_dir, 'concretizer'))

    def get_simulator(self):
        return ScenicSimulator(
            output_dir=os.path.join(self.output_dir, 'simulator'),
            gui=self.settings.get('simulator_settings', {}).get('gui', False),
            seed=self.settings.get('simulator_settings', {}).get('random_seed'))

    def get_monitor(self):
        return ScenicMonitor(os.path.join(self.output_dir, 'monitor'))

# end of class ScenicConfiguration
class CarlaConfiguration(VIVASConfiguration):
    def get_concretizer(self):
        return CarlaConcretizer(
            output_dir=os.path.join(self.output_dir, 'concretizer'))

    def get_simulator(self):
       return CarlaSimulator(
           self.settings.get('simulator_settings', {}).get('host_pipe', False),
           self.settings.get('simulator_settings', {}).get('host_simulator_exe'),
           self.settings.get('simulator_settings', {}).get('scenario_host_location'),
           self.settings['simulator_settings']['output_location'],
           output_dir=os.path.join(self.output_dir, 'simulator'),)

    def get_monitor(self):
        return CarlaMonitor(os.path.join(self.output_dir, 'monitor'))

# end of class ScenicConfiguration


def get_configuration(settings, output_dir):
    backend = settings.get('backend', '3drov').lower()
    if backend == 'scenic':
        return ScenicConfiguration(settings, output_dir)
    elif backend == 'carla':
        return CarlaConfiguration(settings, output_dir)
    else:
        return RoverConfiguration(settings, output_dir)

