import os
from dataclasses import dataclass
from typing import List
try:
    import jstyleson as json
except ImportError:
    import json
import logging


@dataclass
class CoverageScenario:
    initial : List[str]
    intermediate : List[str]
    final : List[str]
    intermediate_ordered : bool

    def __str__(self):
        return json.dumps({'initial' : self.initial,
                           'intermediate' : self.intermediate,
                           'final' : self.final})


class CoverageManager:
    def __init__(self, predicate_map_file, output_dir):
        self.output_dir = output_dir
        self.predicate_map = None
        self.db = {}
        if predicate_map_file:
            with open(predicate_map_file, encoding='utf-8') as f:
                self.predicate_map = json.load(f)

    def get_coverage_scenario(self, abstract_scenario):
        if self.predicate_map is None:
            return None
        res = {}
        logging.info(f'generating coverage scenario for: {abstract_scenario}')
        for phase in ('initial', 'intermediate', 'final'):
            l = []
            for pred in abstract_scenario.get(phase, []):
                cpred = self.predicate_map.get(pred)
                if cpred:
                    l.append(cpred)
            res[phase] = l
        s = CoverageScenario(
            res['initial'], res['intermediate'], res['final'],
            abstract_scenario.get('intermediate_ordered', False))
        return s

    def update_result(self, abstract_scenario, result,
                      concrete_scenario):
        #key = json.dumps(abstract_scenario)
        if concrete_scenario is not None:
            scen = os.path.relpath(concrete_scenario, self.output_dir)
        else:
            scen = None
        if result is None:
            # covdb = self.db.setdefault("coverage", {})
            # l = covdb.setdefault(key, [])
            if scen is not None:
                self.db.setdefault("simulation_failures", []).append(scen)
            return            
        covresult = result.coverage_result
        if covresult is not None:
            # covdb = self.db.setdefault("coverage", {})
            # l = covdb.setdefault(key, [])
            if covresult:
                #l.append(scen)
                self.db.setdefault("coverage", []).append(scen)
            else:
                self.db.setdefault("coverage_failures", []).append(scen)
        propresult = result.property_result
        if propresult != 'OK':
            propdb = self.db.setdefault("property", {})
            propdb.setdefault(propresult, []).append(scen)

    def save(self):
        if self.db:
            outfile = os.path.join(self.output_dir, 'coverage.json')
            with open(outfile, 'w') as out:
                json.dump(self.db, out, indent=2)
                out.write('\n')
                logging.info(f'coverage data saved to {outfile}')

# end of class CoverageManager

