import json
import re
import logging
from .util import get_trace


def _convert_predicate(p, log):
    body = re.sub(r'{([^}]+)}', r'state["\1"]', p)
    try:
        expr = 'lambda state: ' + body
        log(f'converting coverage predicate:\n  {p}\nto expr:\n  {expr}')
        return eval(expr)
    except Exception as e:
        log(f'ERROR in converting coverage predicate: {p},'
            f' assuming always False ({e})')
        return lambda state: False


def _check_coverage(simulation_result, coverage_scenario, logfile):
    if logfile is not None:
        def log(s):
            logfile.write(s)
            logfile.write('\n')
    else:
        def log(s):
            logging.debug(s)
            
    if coverage_scenario is None:
        log('no scenario given')
        return None
    
    initial = coverage_scenario.initial
    intermediate = coverage_scenario.intermediate
    final = coverage_scenario.final
    
    initial_p = [_convert_predicate(p, log) for p in initial]
    intermediate_p = [_convert_predicate(p, log) for p in intermediate]
    final_p = [_convert_predicate(p, log) for p in final]

    states = get_trace(simulation_result)
    log(f'checking coverage on a trace with {len(states)} states')
    
    if not (initial or intermediate or final):
        return None

    def is_true(p, s, state):
        try:
            return p(state)
        except Exception as e:
            log(f'ERROR in evaluating coverage predicate: {s}, '
                f'assuming False ({e})')
            return False

    if not initial:
        log('no initial predicate found')
    for p, s in zip(initial_p, initial):
        log(f'evaluating initial predicate: {s}')
        if not is_true(p, s, states[0]):
            log('violated!')
            return False

    if not final:
        log('no final predicate found')
    for p, s in zip(final_p, final):
        log(f'evaluating final predicate: {s}')
        if not is_true(p, s, states[-1]):
            log('violated!')
            return False

    # TODO -- deal with ordering for intermediates somehow
    if not intermediate:
        log('no intermediate predicate found')
    seen = [False] * len(intermediate)
    if not coverage_scenario.intermediate_ordered:
        for j in range(1, len(states)-1):
            state = states[j]
            for i, (p, s) in enumerate(zip(intermediate_p, intermediate)):
                if not seen[i] and is_true(p, s, state):
                    log(f'intermediate predicate {i}:\n  {s}'
                        f'\nholds at state {j}')
                    seen[i] = True
    else:
        i = 0
        for j in range(1, len(states)-1):
            state = states[j]
            p, s = intermediate_p[i], intermediate[i]
            if is_true(p, s, state):
                log(f'intermediate predicate {i}:\n  {s}'
                    f'\nholds at state {j}')
                seen[i] = True
                i += 1
                if i >= len(intermediate):
                    break

    if not all(seen):
        log('violated intermediate predicates:\n  ' +
            '\n  '.join(s for (i, s) in enumerate(intermediate)
                        if not seen[i]))
        return False
    else:
        return True


def check_coverage(simulation_result, coverage_scenario, output_name):
    try:
        logfile = None
        if output_name is not None:
            logfile = open(output_name, 'w')
        return _check_coverage(simulation_result, coverage_scenario, logfile)
    finally:
        if logfile is not None:
            logfile.close()
    

if __name__ == '__main__':
    import sys
    logging.getLogger().setLevel(logging.DEBUG)
    trace = sys.argv[1]
    predmap = json.load(open(sys.argv[2]))
    prop = json.load(open(sys.argv[3]))
    class ScenarioStub:
        def __init__(self, d):
            def fill(key):
                res = []
                for p in d.get(key, []):
                    pm = predmap.get(p)
                    if pm:
                        res.append(pm)
                return res
            self.initial = fill("initial")
            self.intermediate = fill("intermediate")
            self.final = fill("final")
            self.intermediate_ordered = d.get("intermediate_ordered", False)

    scen = ScenarioStub(prop)
    res = check_coverage(trace, scen, None)
    print(res)

