import shutil
import os
from simulator.interfaces import SimulationResult, AbstractSimulator


class CarlaSimulationResult(SimulationResult):
    def __init__(self, is_error, trace_file):
        self.is_error = is_error
        self.trace_file = trace_file

        pass
    def error(self):
        return self.is_error
    
    def to_monitor_args(self):
        raise NotImplementedError

# end of class RoverSimulationResult


class CarlaSimulator(AbstractSimulator):

    """ Class used to execute a single instance of a Carla simulation
    directly on host. """
    def __init__(self, docker_pipe: str, simulator_exe: str, scenario_host_location: str, output_location: str, output_dir):
        """
        Constructor for class RoverSimulatorInterface
        Parameters
        ----------
        simulator_exe
            Path to the simulator executable to be used to run simulations.

        docker_pipe
            Location for the docker pipe to run commands (in this case the
            simulation) on the host

        """

        self._simulator_exe = simulator_exe
        self._docker_pipe = docker_pipe
        self._scenario_host_location = scenario_host_location
        self._output_location = output_location
        self._output_dir = output_dir


    def run_simulation(self, scenario_file: str, ego_route_file: str):

        # Load the scenario into a string that can be piped to the host
        # Move the scenario file to where it should run
        shutil.copy(scenario_file, os.path.join(self._scenario_host_location, "vivas_scenario.py"))
        shutil.copy(ego_route_file, os.path.join(self._scenario_host_location, "ego_route.xml"))
        print(f"opening pipe: {self._docker_pipe}")

        # Ugly hack to run the simulation on HOST while we are inside the
        # docker. Requires that the simulator server is running on the host and
        # that a named pipe is connect to inside the docker using the -v flag.
        # pipe.write() will execute the argument in a terminal on the host system. 
        # For the named pipe stuff, see:
        # https://stackoverflow.com/questions/32163955/how-to-run-shell-script-on-host-from-docker-container
        with open(self._docker_pipe, 'w') as pipe:

            # Run the simulator script
            command = self._simulator_exe + '\n'
            print("RUNNING COMMAND:")
            print(command)
            pipe.write(command)

        with open(self._docker_pipe, 'r') as pipe:

            while pipe.readline() != "Simulation finished\n":
                pass

        print("Simulation finished")
        
        trace = os.path.join(self._output_location, "carla_result.json")
        bn = os.path.splitext(os.path.basename(scenario_file))[0]
        trace_file = shutil.copy(trace, os.path.join(self._output_dir, bn + '-trace.json'))



        return CarlaSimulationResult(False, trace_file)


    def cleanup(self, concrete_scenario: str):
        pass

    def retrieve_simulation(self, concrete_scenario):
        name = os.path.splitext(os.path.basename(concrete_scenario))[0]
        trace_file = os.path.join(self._output_dir, name + '-trace.json')
        if os.path.exists(trace_file):
            return CarlaSimulationResult(False, trace_file)
        else:
            return CarlaSimulationResult(True, None)

# end of class CarlaSimulator

