import argparse
import logging
import os
import shutil
import subprocess
import sys
import config
import random
import glob
try:
    import jstyleson as json
except ImportError:
    import json

green = '32'
yellow = '33'
cyan = '36'
red = '31'
blue = '34'
magenta = '95'

def color(s, c):
    return f'\033[1;{c}m{s}\033[0m' if os.isatty(sys.stdout.fileno()) else s


def banner(msg, textcolor=None):
    print('-' * 77)
    if textcolor:
        print(color(msg, textcolor))
    else:
        print(msg)
    print('-' * 77)


def getopts():
    parser = argparse.ArgumentParser()
    parser.add_argument("config_file",
                        help="Path to file containing "
                        "the necessary configurations")
    parser.add_argument('output_dir',
                        help='output directory to store the results')
    parser.add_argument('-f', '--force-overwrite', action='store_true',
                        help='force overwriting the output directory')
    parser.add_argument('-m', '--mode', choices=['full', 'replay'],
                        default='full')
    args = parser.parse_args()
    args.output_dir = os.path.normpath(os.path.abspath(args.output_dir))
    return args


def get_abstract_scenarios(opts, inputfile, absgen):
    if opts.mode == 'full':
        yield from absgen.generate_abstract_scenarios(inputfile)
    else:
        for name in glob.glob(os.path.join(opts.output_dir,
                                           'absgen', 'properties_*.json')):
            idx = os.path.splitext(os.path.basename(name))[0].split('_', 1)[1]
            trace = os.path.join(opts.output_dir, 'absgen',
                                 f'scenario_{idx}.xml')
            if os.path.exists(trace):
                with open(name, encoding='utf-8') as f:
                    property_dict = json.load(f)
                yield (trace, property_dict)


def get_concrete_scenarios(opts, concretizer, abstract_trace, abstract_scenario):
    if opts.mode == 'full':
        yield from concretizer.create_scenarios(abstract_trace,
                                                abstract_scenario)
    else:
        yield from concretizer.retrieve_scenarios(abstract_trace,
                                                  abstract_scenario)


def get_simulation_result(opts, simulator, concrete_scenario):
    if opts.mode == 'full':
        return simulator.run_simulation(concrete_scenario[0], concrete_scenario[1])
    else:
        return simulator.retrieve_simulation(concrete_scenario)


def vivas_loop(opts, inputfile, absgen, concretizer, simulator, monitor, covmgr):
    print(color('=' * 77, cyan))
    for absres in get_abstract_scenarios(opts, inputfile, absgen):
        abstract_trace, abstract_scenario = absres
        if abstract_trace is None:
            banner('ERROR in generating abstract scenario', red)
            print(color('=' * 77, cyan))
            covmgr.update_result(abstract_scenario, None, None)
            covmgr.save()
            continue
        banner(f'got abstract scenario: {abstract_trace}', green)
        for concrete_scenario in get_concrete_scenarios(opts, concretizer,
                                                        abstract_trace,
                                                        abstract_scenario):
            banner(f"running concrete scenario {concrete_scenario} in simulator",
                   yellow)
            print(concrete_scenario[0])
            result = get_simulation_result(opts, simulator, concrete_scenario)

            if result.error():
                banner(f'ERROR in simulating the concrete scenario '
                       f'{concrete_scenario}', red)
                covmgr.update_result(abstract_scenario, None, concrete_scenario[0])
            else:
                banner(f'monitoring simulation run for {concrete_scenario[0]}', red)

                scenario_name = os.path.splitext(
                    os.path.basename(concrete_scenario[0]))[0]
                coverage_scenario = covmgr.get_coverage_scenario(
                    abstract_scenario)
                monitor_result = monitor.run(scenario_name, coverage_scenario,
                                             result)
                property_result = monitor_result.property_result
                c = yellow
                if property_result.startswith('ERROR'):
                    c = magenta
                elif property_result == 'OK':
                    c = green
                elif property_result == 'FAIL':
                    c = red
                if monitor_result.coverage_result is None:
                    covres = color('N/A', yellow)
                elif monitor_result.coverage_result:
                    covres = color('YES', green)
                else:
                    covres = color('NO', red)
                    if property_result == 'FAIL':
                        property_result = 'WARNING'
                        c = yellow
                print(f'RESULT: {color(property_result, c)}')
                print(f'COVERAGE RESULT: {covres}')

                covmgr.update_result(abstract_scenario, monitor_result,
                                     concrete_scenario[0])

            simulator.cleanup(concrete_scenario)
        covmgr.save()
        print(color('=' * 77, cyan))


def main():
    random.seed(42)
    logging.basicConfig(level=logging.INFO)

    args = getopts()

    if args.mode == 'full':
        if os.path.exists(args.output_dir):
            if args.force_overwrite:
                shutil.rmtree(args.output_dir)
            else:
                banner(f'ERROR: output directory "{args.output_dir}" '
                       'already exists!', red)
                return 1
        os.makedirs(args.output_dir)
        os.mkdir(os.path.join(args.output_dir, 'absgen'))
        os.mkdir(os.path.join(args.output_dir, 'concretizer'))
        os.mkdir(os.path.join(args.output_dir, 'simulator'))
        os.mkdir(os.path.join(args.output_dir, 'monitor'))
    else:
        for d in ('.', 'absgen', 'concretizer', 'simulator', 'monitor'):
            if not os.path.exists(os.path.join(args.output_dir, d)):
                banner(f'ERROR: results directory "{args.output_dir}" '
                       'does not exist or is malformed!', red)
                return 1

    # Parse input settings
    settings = None
    with open(args.config_file, "r", encoding="utf-8") as settings_file:
        settings = json.load(settings_file)
        settings['basedir'] = os.path.abspath(os.path.dirname(args.config_file))

    conf = config.get_configuration(settings, args.output_dir)
    
    # instantiate everything
    absgen = conf.get_abstract_generator()
    concretizer = conf.get_concretizer()
    simulator = conf.get_simulator()
    monitor = conf.get_monitor()
    covmgr = conf.get_coverage_manager()
    
    concrete_scenarios_basenames = []

    info = f'*** VIVAS DEMO ***\n' \
        f'conf_file = {args.config_file}\n' \
        f'output_dir = {args.output_dir}'
    banner(info, blue)
    banner('beginning enumeration', cyan)

    ag_set = settings["absgen_settings"]
    inputfiles = ag_set["input_file"]
    if isinstance(inputfiles, str):
        inputfiles = [inputfiles]
    for inputfile in inputfiles:
        if not os.path.isabs(inputfile):
            inputfile = os.path.abspath(os.path.join(settings['basedir'],
                                                      inputfile))
        vivas_loop(args,
                   inputfile, absgen, concretizer, simulator, monitor, covmgr)

    return 0


if __name__ == "__main__":
    main()

